#ifndef _FL_OPDESK_H
#define _FL_OPDESK_H

//////////////////////
// Fl_OpDesk.H
//////////////////////
//
// OpDesk (Version 0.70)
// This file is part of the OpDesk node graph FLTK widget.
// Copyright (C) 2009 by Greg Ercolano.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//

///
/// \file
/// \brief This file contains the definitions for the Fl_OpDesk widget.
///

#include <FL/Fl.H>
#include <FL/Fl_Scroll.H>
#include <FL/fl_draw.H>

#include <vector>		// STL vector array

#include "Fl_OpBox.H"
#include "Fl_OpConnect.H"

/// \enum Fl_OpConnectStyle
/// Style for the connection lines between box buttons
///
enum Fl_OpConnectStyle {
    FL_OPCONNECT_STYLE_STRAIGHT,	///< -- 'Single' style line connections between buttons
    FL_OPCONNECT_STYLE_TRACE,		///< -- 'Solder trace' style line connections
};

/// \class Fl_OpDesk
///
/// \brief The main Fl_OpDesk widget, parent to the following widgets:
///      Fl_OpBox -- the boxes that can be connected together
///      Fl_OpButton -- the buttons (input and output) that can be added to an Fl_OpBox
///      Fl_OpConnect -- the connections between input/output buttons of different boxes.
///
/// \code
///
///     Fl_OpDesk                     // Desk -- top level widget that contains all boxes
///        |___ Fl_OpBox              // Boxes -- the operation boxes
///               |___ Fl_OpButton    // Buttons -- the buttons in the boxes
///
/// \endcode
///
class Fl_OpDesk : public Fl_Scroll {

    int opconnect_spacing;		// desk's spacing between opconnect lines
    Fl_OpConnectStyle opconnect_style;	// desk's line style (straight,trace) for opconnect lines
    int opconnect_width;		// default line width for new opconnect's
    Fl_Color opconnect_color;		// default color used for new opconnect's
    std::vector<Fl_OpConnect> connects;	// our array of all button interconnects
    std::vector<Fl_OpBox*> boxes;		// boxes on desk

    // Clipboard
    std::vector<Fl_OpBox*> box_clipboard;		// copy/paste clipboard for Fl_OpBox's

    // Connection dragging
    char dragging;			// on/off flag: 1=dragging, 0=not
    int dragging_x1, dragging_y1,	// when dragging=1, these contain the two
        dragging_x2, dragging_y2;	// points to draw a connection line between.

protected:
    // PROTECTED METHODS
    // Drawing..
    void DrawLine(float x1, float y1, float x2, float y2) const;
    void DrawConnect(Fl_OpConnect *con) const;

    // Connections..
    friend class Fl_OpButton;
    friend class Fl_OpBox;
    int _ConnectOnly(Fl_OpButton *srcbut, Fl_OpButton *dstbut, std::string &errmsg);
    int _DisconnectOnly(Fl_OpButton *a, Fl_OpButton *b);
    void CopyConnections(std::vector<Fl_OpBox*> origboxes, std::vector<Fl_OpBox*> copyboxes);

    // Dragging
    void DraggingConnection(Fl_OpButton *b);
    void DraggingBoxes(int xdiff, int ydiff);

    void _AddBox(Fl_OpBox *b);
    void _RemoveBox(Fl_OpBox *b);

public:
    // CTOR
    Fl_OpDesk(int X,int Y,int W,int H,const char *L=0);
    ~Fl_OpDesk();

    // FLTK
    void draw();
    int handle(int e);

    // OPBOX
    void DeleteBox(Fl_OpBox *box);
    int GetOpBoxIndex(const Fl_OpBox*) const;
    int GetOpBoxTotal() const;
    Fl_OpBox* GetOpBox(int index);
    const Fl_OpBox* GetOpBox(int t) const;

    // SELECTION
    int DeselectAll();
    int SelectAll();

    // COPY/PASTE
    virtual int DeleteSelected();
    virtual int CutSelected();
    virtual int CopySelected();
    virtual int PasteSelected();

    // SEARCH FUNCTIONS
    Fl_OpBox*    FindBoxForFullName(const std::string& fullname, std::string &errmsg);
    Fl_OpButton* FindButtonForFullName(const std::string& fullname, std::string &errmsg);
    Fl_OpBox*    FindBoxByLabel(std::string &lname, std::string &errmsg);
    Fl_OpButton* FindButtonUnderMouse();

    // MAKE/DELETE CONNECTIONS, CONNECTION ATTRIBUTES
    int GetOpConnectSpacing() const;
    void SetOpConnectSpacing(int val);
    int GetOpConnectWidth() const;
    void SetOpConnectWidth(int val);
    Fl_Color GetOpConnectColor() const;
    void SetOpConnectColor(Fl_Color val);
    int GetConnectionsTotal() const;
    Fl_OpConnect* GetConnection(int index);
    Fl_OpConnect* GetConnection(Fl_OpButton *srcbut, Fl_OpButton *dstbut);
    Fl_OpConnect* GetLastOpConnect();
    int Connect(Fl_OpButton *srcbut, Fl_OpButton *dstbut, std::string &errmsg);
    int Connect(Fl_OpBox *srcbox, const std::string& srcbut_lname, 
	        Fl_OpBox *dstbox, const std::string& dstbut_lname,
	        std::string &errmsg);
    int Connect(const std::string& src_name, 
		const std::string& dst_name,
		      std::string& errmsg);
    int Disconnect(Fl_OpButton *a, Fl_OpButton *b);
    int Disconnect(Fl_OpBox *box);
    void DisconnectAll();
    void BringToFront(Fl_OpBox *box);

    // MISC
    void Clear();
    int ParseFullButtonName(const std::string &fullname,
				  std::string &boxname,
				  std::string &butname,
				  std::string &errmsg);

    // VIRTUALS

    /// This virtual method can be defined by the app to define what to do
    /// when a connection error is detected. Default behavior is to print
    /// the \p errmsg to stderr, but app may want to post a dialog instead.
    /// The \p srcbut and \p dstbut are the buttons whose connection attempt had failed.
    ///
    /// Example:
    /// \code
    /// class MyDesk : public Fl_OpDesk {
    ///     // Handle displaying connection errors
    ///     void ConnectionError(Fl_OpButton *src, Fl_OpButton *dst, std::string &errmsg) {
    ///         std::string msg;
    /// 	msg = std::string("    INPUT: ") + src->GetFullName() + std::string("\n") +
    /// 	      std::string(   "OUTPUT: ") + dst->GetFullName() + std::string("\n") +
    /// 	      errmsg;
    ///         fl_alert(msg.c_str());		// post a dialog
    ///     }
    /// [..]
    /// \endcode
    ///
    virtual void ConnectionError(Fl_OpButton* /*srcbut*/, Fl_OpButton* /*dstbut*/, std::string &errmsg) {
        fprintf(stderr, "CONNECTION ERROR: %s\n", errmsg.c_str());
    }
};

#endif /*_FL_OPDESK_H*/
